# Generated by Django 3.2.9 on 2021-11-16 17:26

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('rgd', '0005_auto_20211105_1715'),
        ('rgd_3d', '0002_alter_pointcloud_status'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='PointCloud',
            new_name='Mesh3D',
        ),
        migrations.RenameModel(
            old_name='PointCloudMeta',
            new_name='Mesh3DMeta',
        ),
        migrations.RenameModel(
            old_name='PointCloudSpatial',
            new_name='Mesh3DSpatial',
        ),
    ]



import json
import os
import tempfile
from typing import Union

from celery.utils.log import get_task_logger
from django.contrib.gis.geos import Polygon
import numpy as np
import pyproj
from rgd.models import DB_SRID, ChecksumFile
from rgd.utility import get_or_create_no_commit, get_temp_dir
from rgd_3d.models import Mesh3D, Tiles3D, Tiles3DMeta

logger = get_task_logger(__name__)


def _file_conversion_helper(source, output_field, method, prefix='', extension='', **kwargs):
    workdir = get_temp_dir()
    with tempfile.TemporaryDirectory(dir=workdir) as tmpdir:
        # NOTE: cannot use FUSE with PyntCloud because it checks file extension
        #       in path. Additionally, The `name` field for the file MUST have
        #       the extension.
        with source.yield_local_path(try_fuse=False) as file_path:
            output_path = os.path.join(tmpdir, prefix + os.path.basename(source.name) + extension)
            method(str(file_path), str(output_path), **kwargs)
        with open(output_path, 'rb') as f:
            output_field.save(os.path.basename(output_path), f)


def _save_pyvista(mesh, output_path):
    surf = mesh.extract_surface()
    surf.save(output_path)


def _use_pyntcloud_pyvista(input_path, output_path):
    from pyntcloud import PyntCloud

    cloud = PyntCloud.from_file(input_path)
    mesh = cloud.to_instance('pyvista', mesh=False)
    _save_pyvista(mesh, output_path)


def _use_pyvista(input_path, output_path):
    import pyvista as pv

    mesh = pv.read(input_path)
    _save_pyvista(mesh, output_path)


def _get_readers():
    import pyvista as pv

    methods = {
        'las': _use_pyntcloud_pyvista,
        'npy': _use_pyntcloud_pyvista,
        'npz': _use_pyntcloud_pyvista,
    }
    methods.update({k[1::]: _use_pyvista for k in pv.utilities.fileio.READERS.keys()})
    return methods


def read_mesh_3d_file(mesh_3d: Union[Mesh3D, int]):
    """Read a Mesh3D object and populate the vtp_data field."""
    if not isinstance(mesh_3d, Mesh3D):
        mesh_3d = Mesh3D.objects.get(pk=mesh_3d)
    if not mesh_3d.name:
        mesh_3d.name = mesh_3d.file.name
    # Parse the point cloud file format and convert
    ext = mesh_3d.file.name.split('.')[-1].strip().lower()
    try:
        method = _get_readers()[ext]
    except KeyError:
        raise ValueError(f'Extension `{ext}` unsupported for point cloud conversion.')
    if not mesh_3d.vtp_data:
        mesh_3d.vtp_data = ChecksumFile()
    if not mesh_3d.vtp_data.collection:
        mesh_3d.vtp_data.collection = mesh_3d.file.collection
    _file_conversion_helper(mesh_3d.file, mesh_3d.vtp_data.file, method, extension='.vtp')
    # Save the record
    mesh_3d.save(
        update_fields=[
            'name',
            'vtp_data',
        ]
    )


def read_3d_tiles_tileset_json(tiles_3d: Union[Tiles3D, int]):
    if not isinstance(tiles_3d, Tiles3D):
        tiles_3d = Tiles3D.objects.get(pk=tiles_3d)
    tileset_json = json.load(tiles_3d.json_file.file)

    tiles_3d_meta, created = get_or_create_no_commit(Tiles3DMeta, source=tiles_3d)

    volume = tileset_json['root']['boundingVolume']
    logger.debug(f'3D tiles bounding volume: {volume}')

    def bounds_to_polygon(xmin, xmax, ymin, ymax):
        return np.array(
            [
                (xmin, ymax),
                (xmax, ymax),
                (xmax, ymin),
                (xmin, ymin),
                (xmin, ymax),  # close the loop
            ]
        )

    transform = np.array(tileset_json['root']['transform']).reshape((4, 4), order='C')

    # The coordinates should be in EPSG:4978 by convention in 3D Tiles
    # See https://github.com/CesiumGS/3d-tiles/tree/main/specification#coordinate-reference-system-crs
    # srid = 4978
    if 'region' in volume:
        xmin, ymin, xmax, ymax, _, _ = volume['region']
        # The region bounding volume specifies bounds using a geographic coordinate system (latitude, longitude, height), specifically EPSG 4979.
        # See https://github.com/CesiumGS/3d-tiles/tree/main/specification#coordinate-reference-system-crs
        # srid = 4979
        coords = bounds_to_polygon(xmin, xmax, ymin, ymax) * 180.0 / np.pi
    elif 'box' in volume:
        b = volume['box']
        # NOTE: I don't think this is right
        center, x, y, _ = b[0:3], b[3:6], b[6:9], b[9:12]
        xmin = min(center[0], x[0], y[0])
        xmax = max(center[0], x[0], y[0])
        ymin = min(center[1], x[1], y[1])
        ymax = max(center[1], x[1], y[1])
        coords = bounds_to_polygon(xmin, xmax, ymin, ymax)
        coords = np.c_[coords, np.zeros(len(coords)), np.ones(len(coords))]
        # Apply the tranformation
        corners = (coords @ transform)[:, :-1]  # in XYZ world
        src = pyproj.CRS('EPSG:4978')  # 4979 -> srid
        dest = pyproj.CRS('EPSG:4326')
        transformer = pyproj.Transformer.from_proj(src, dest, always_xy=True)
        x, y, _ = transformer.transform(corners.T[0], corners.T[1], corners.T[2])
        coords = np.c_[x, y]
    elif 'sphere' in volume:
        raise NotImplementedError
    else:
        raise ValueError(f'Bounding volume of unknown type: {volume}')

    outline = Polygon(coords, srid=DB_SRID)

    # Save out
    tiles_3d_meta.outline = outline
    tiles_3d_meta.footprint = outline  # TODO: handle oriented bounding boxes

    tiles_3d_meta.save()



from django.contrib.gis.db import models
from django.contrib.postgres import fields
from django_extensions.db.models import TimeStampedModel
from rgd.models import ChecksumFile, SpatialEntry
from rgd.models.mixins import DetailViewMixin, TaskEventMixin
from rgd_3d.tasks import jobs


class Mesh3D(TimeStampedModel, TaskEventMixin, DetailViewMixin):
    """Container for point cloud file."""

    name = models.CharField(max_length=1000, blank=True)
    description = models.TextField(null=True, blank=True)

    # Source data uploaded by user in just about any format
    file = models.ForeignKey(ChecksumFile, on_delete=models.CASCADE, related_name='+')

    # A place to store converted file - must be in VTP format
    vtp_data = models.ForeignKey(
        ChecksumFile, on_delete=models.DO_NOTHING, related_name='+', null=True, blank=True
    )

    def data_link(self):
        return self.file.data_link()

    def data_link_vtp(self):
        return self.vtp_data.data_link()

    data_link.allow_tags = True
    data_link_vtp.allow_tags = True

    task_funcs = (jobs.task_read_mesh_3d_file,)
    detail_view_name = 'detail-mesh-3d'


class Mesh3DSpatial(TimeStampedModel, SpatialEntry):
    """Optionally register a Mesh3D as a SpatialEntry."""

    source = models.OneToOneField(Mesh3D, on_delete=models.CASCADE)

    crs = models.TextField(help_text='PROJ string', blank=True, null=True)  # PROJ String
    # Origin point to map the 0,0,0 point of the point cloud
    origin = fields.ArrayField(models.FloatField(), size=3, blank=True, null=True)

    @property
    def name(self):
        return self.source.file.name

    detail_view_name = Mesh3D.detail_view_name
    detail_view_pk = 'source__pk'



from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from rest_framework.decorators import action
from rest_framework.generics import get_object_or_404
from rest_framework.request import Request
from rgd.models.file import ChecksumFile
from rgd.rest.base import ModelViewSet
from rgd.rest.mixins import TaskEventViewSetMixin
from rgd.utility import get_file_data_url
from rgd_3d import models, serializers


class Mesh3DViewSet(ModelViewSet, TaskEventViewSetMixin):
    serializer_class = serializers.Mesh3DSerializer
    queryset = models.Mesh3D.objects.all()

    @action(
        detail=True,
        serializer_class=serializers.Mesh3DDataSerializer,
    )
    def base64(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)


class Tiles3DViewSet(ModelViewSet):
    queryset = models.Tiles3D.objects.all()

    def get_serializer_class(self):
        if self.action in {'list', 'retrieve', 'status'}:
            return serializers.Tiles3DMetaSerializer
        return serializers.Tiles3DSerializer

    def get_queryset(self):
        if self.action in {'list', 'retrieve', 'status'}:
            return models.Tiles3DMeta.objects.all()
        return models.Tiles3D.objects.all()

    @swagger_auto_schema(responses={302: openapi.Response('Redirect to file download')})
    @action(detail=True, methods=['GET'], url_path='file/(?P<name>.+)')
    def file(self, request: Request, pk: str, name: str):
        """Download a file from a 3D tiles FileSet."""
        tiles3d_set: models.Tiles3D = get_object_or_404(models.Tiles3D, pk=pk)
        checksum_file: ChecksumFile = get_object_or_404(
            tiles3d_set.json_file.file_set.files, name=name
        )
        return get_file_data_url(checksum_file)



from typing import List

from pooch.processors import Unzip
from rgd.datastore import datastore
from rgd.management.commands._data_helper import SynchronousTasksCommand, _get_or_create_file_model
from rgd_3d import models
from rgd_3d.models.tiles import create_tiles3d_from_paths

SUCCESS_MSG = 'Finished loading all demo data.'

# Names of files in the datastore
POINT_CLOUD_FILES = ['topo.vtk']
TILES_3D_FILES = ['jacksonville-untextured.zip', 'jacksonville-textured.zip', 'dragon.zip']


def load_mesh_3d_files(pc_files: List[str]):
    ids = []
    for f in pc_files:
        entry = _get_or_create_file_model(models.Mesh3D, f)
        ids.append(entry.pk)
    return ids


def load_tiles_3d_files(t3d_files: List[str]):
    ids = []
    for f in t3d_files:
        paths = datastore.fetch(f, processor=Unzip())
        entry = create_tiles3d_from_paths(paths)
        ids.append(entry.pk)
    return ids


class Command(SynchronousTasksCommand):
    help = 'Populate database with demo data.'

    def handle(self, *args, **options):

        self.set_synchronous()
        # Run the command
        load_mesh_3d_files(POINT_CLOUD_FILES)

        load_tiles_3d_files(TILES_3D_FILES)

        self.stdout.write(self.style.SUCCESS(SUCCESS_MSG))
        self.reset_celery()



# Generated by Django 3.2.7 on 2021-09-27 15:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rgd_3d', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pointcloud',
            name='status',
            field=models.CharField(
                choices=[
                    ('created', 'Created but not queued'),
                    ('queued', 'Queued for processing'),
                    ('running', 'Processing'),
                    ('failed', 'Failed'),
                    ('success', 'Succeeded'),
                    ('skipped', 'Skipped'),
                ],
                default='created',
                max_length=20,
            ),
        ),
    ]



from pooch.processors import Unzip
import pytest
from rgd.datastore import datastore
from rgd_3d.models.tiles import Tiles3DMeta, create_tiles3d_from_paths

TOLERANCE = 2e-2

centroids = {
    'jacksonville-untextured.zip': {'x': -81.6634, 'y': 30.3234},
    'jacksonville-textured.zip': {'x': -81.6634, 'y': 30.3234},
    'dragon.zip': {'x': -75.6079, 'y': 40.0439},
}


@pytest.mark.parametrize(
    'sample_file',
    [
        'jacksonville-untextured.zip',
        'jacksonville-textured.zip',
        'dragon.zip',
    ],
)
@pytest.mark.django_db(transaction=True)
def test_mesh_3d_etl(sample_file):
    paths = datastore.fetch(sample_file, processor=Unzip())

    entry = create_tiles3d_from_paths(paths)
    entry.refresh_from_db()

    meta = Tiles3DMeta.objects.get(source=entry)

    assert meta.footprint is not None
    centroid = meta.outline.centroid
    assert centroid.x == pytest.approx(centroids[sample_file]['x'], abs=TOLERANCE)
    assert centroid.y == pytest.approx(centroids[sample_file]['y'], abs=TOLERANCE)



from django.urls import path
from rest_framework.routers import SimpleRouter
from rgd_3d import models, views
from rgd_3d.rest import viewsets

router = SimpleRouter(trailing_slash=False)
router.register(r'api/rgd_3d/mesh', viewsets.Mesh3DViewSet, basename='mesh-3d')
router.register(r'api/rgd_3d/tiles3d', viewsets.Tiles3DViewSet, basename='tiles-3d')

urlpatterns = [
    # Pages
    path(
        'rgd_3d/mesh/<int:pk>/',
        views.Mesh3DDetailView.as_view(),
        name=models.Mesh3D.detail_view_name,
    ),
    path(
        'rgd_3d/tiles3d/<int:pk>/',
        views.Tiles3DDetailView.as_view(),
        name=models.Tiles3D.detail_view_name,
    ),
] + router.urls



import pytest
from rest_framework import status


@pytest.mark.django_db(transaction=True)
def test_swagger(admin_api_client):
    response = admin_api_client.get('/swagger/?format=openapi')
    assert status.is_success(response.status_code)



from celery import shared_task
from django.conf import settings
from rgd.tasks import helpers


@shared_task(time_limit=settings.CELERY_TASK_TIME_LIMIT)
def task_read_mesh_3d_file(pc_file_pk):
    from rgd_3d.models import Mesh3D

    from .etl import read_mesh_3d_file

    pc_file = Mesh3D.objects.get(pk=pc_file_pk)
    helpers._run_with_failure_reason(pc_file, read_mesh_3d_file, pc_file_pk)


@shared_task(time_limit=settings.CELERY_TASK_TIME_LIMIT)
def task_read_3d_tiles_file(tiles_3d_pk: int):
    from rgd_3d.models import Tiles3D

    from .etl import read_3d_tiles_tileset_json

    tiles_3d = Tiles3D.objects.get(pk=tiles_3d_pk)
    helpers._run_with_failure_reason(tiles_3d, read_3d_tiles_tileset_json, tiles_3d_pk)



# Generated by Django 3.2.9 on 2021-11-16 18:14

from django.db import migrations, models
import django.db.models.deletion


def migrate_mesh3dmeta_to_mesh3d(apps, schema_editor):
    Mesh3DMeta = apps.get_model('rgd_3d', 'Mesh3DMeta')  # noqa
    for meta in Mesh3DMeta.objects.all():
        meta.source.name = meta.name
        meta.source.description = meta.description
        meta.source.vtp_data = meta.vtp_data
        meta.source.save(
            update_fields=[
                'name',
                'description',
                'vtp_data',
            ]
        )


class Migration(migrations.Migration):

    dependencies = [
        ('rgd', '0005_auto_20211105_1715'),
        ('rgd_3d', '0003_auto_20211116_1726'),
    ]

    operations = [
        migrations.AddField(
            model_name='mesh3d',
            name='description',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='mesh3d',
            name='name',
            field=models.CharField(blank=True, max_length=1000),
        ),
        migrations.AddField(
            model_name='mesh3d',
            name='vtp_data',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name='+',
                to='rgd.checksumfile',
            ),
        ),
        migrations.RunPython(migrate_mesh3dmeta_to_mesh3d),
        migrations.DeleteModel(
            name='Mesh3DMeta',
        ),
    ]



from rgd.rest.mixins import BaseRestViewMixin
from rgd.views import PermissionDetailView, PermissionListView, PermissionTemplateView
from rgd_3d.urls import urlpatterns


def test_urls():
    for pattern in urlpatterns:
        if hasattr(pattern.callback, 'view_class') and 'WrappedAPIView' not in str(
            pattern.callback
        ):
            assert issubclass(
                pattern.callback.view_class,
                (
                    BaseRestViewMixin,
                    PermissionDetailView,
                    PermissionTemplateView,
                    PermissionListView,
                ),
            )



# Generated by Django 3.2.9 on 2021-12-07 02:53

from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields
import rgd.models.mixins


class Migration(migrations.Migration):

    dependencies = [
        ('rgd', '0005_auto_20211105_1715'),
        ('rgd_3d', '0004_auto_20211116_1814'),
    ]

    operations = [
        migrations.CreateModel(
            name='Tiles3D',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name='ID'
                    ),
                ),
                ('failure_reason', models.TextField(null=True)),
                (
                    'status',
                    models.CharField(
                        choices=[
                            ('created', 'Created but not queued'),
                            ('queued', 'Queued for processing'),
                            ('running', 'Processing'),
                            ('failed', 'Failed'),
                            ('success', 'Succeeded'),
                            ('skipped', 'Skipped'),
                        ],
                        default='created',
                        max_length=20,
                    ),
                ),
                (
                    'created',
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name='created'
                    ),
                ),
                (
                    'modified',
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name='modified'
                    ),
                ),
                ('name', models.CharField(blank=True, max_length=1000)),
                ('description', models.TextField(blank=True, null=True)),
                (
                    'json_file',
                    models.ForeignKey(
                        help_text='The `tileset.json` file that contains metadata about this set of 3D tiles.',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='+',
                        to='rgd.checksumfile',
                    ),
                ),
            ],
            options={
                'verbose_name': '3D tiles',
                'verbose_name_plural': '3D tiles',
            },
            bases=(
                models.Model,
                rgd.models.mixins.PermissionPathMixin,
                rgd.models.mixins.DetailViewMixin,
            ),
        ),
        migrations.CreateModel(
            name='Tiles3DMeta',
            fields=[
                (
                    'spatialentry_ptr',
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to='rgd.spatialentry',
                    ),
                ),
                (
                    'created',
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name='created'
                    ),
                ),
                (
                    'modified',
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name='modified'
                    ),
                ),
                (
                    'source',
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE, to='rgd_3d.tiles3d'
                    ),
                ),
            ],
            options={
                'get_latest_by': 'modified',
                'abstract': False,
            },
            bases=('rgd.spatialentry', models.Model, rgd.models.mixins.PermissionPathMixin),
        ),
    ]



from rgd.views import PermissionDetailView
from rgd_3d import models


class Mesh3DDetailView(PermissionDetailView):
    model = models.Mesh3D


class Tiles3DDetailView(PermissionDetailView):
    model = models.Tiles3D



import pytest
from rgd.datastore import datastore
from rgd_3d.tasks.etl import read_mesh_3d_file

from . import factories


@pytest.mark.parametrize(
    'sample_file',
    [
        'topo.vtk',
    ],
)
@pytest.mark.django_db(transaction=True)
def test_mesh_3d_etl(sample_file):
    entry = factories.Mesh3DFactory(
        file__file__filename=sample_file,
        file__file__from_path=datastore.fetch(sample_file),
    )
    entry.refresh_from_db()
    assert entry.vtp_data is not None
    # Testing that we can repopulate a point cloud entry
    read_mesh_3d_file(entry.id)



from django.contrib import admin
from django.contrib.gis.admin import OSMGeoAdmin
from rgd.admin.mixins import (
    MODIFIABLE_FILTERS,
    TASK_EVENT_FILTERS,
    TASK_EVENT_READONLY,
    GeoAdminInline,
    reprocess,
)
from rgd_3d.models import Mesh3D, Mesh3DSpatial


class Mesh3DSpatialInline(GeoAdminInline):
    model = Mesh3DSpatial
    fk_name = 'source'
    list_display = (
        'pk',
        'modified',
        'created',
        'crs',
    )
    readonly_fields = (
        'modified',
        'created',
    )
    modifiable = False  # To still show the footprint and outline


@admin.register(Mesh3D)
class Mesh3DAdmin(OSMGeoAdmin):
    list_display = (
        'pk',
        'name',
        'status',
        'modified',
        'created',
        # 'data_link',
        # 'data_link_vtp',
    )
    readonly_fields = (
        'modified',
        'created',
    ) + TASK_EVENT_READONLY
    actions = (reprocess,)
    list_filter = MODIFIABLE_FILTERS + TASK_EVENT_FILTERS
    inlines = (Mesh3DSpatialInline,)
    extra = 0
    raw_id_fields = ('file', 'vtp_data')



import base64

from rest_framework import serializers
from rgd.serializers import ChecksumFileSerializer, SpatialEntryFootprintSerializer
from rgd_3d import models


class Mesh3DSerializer(serializers.ModelSerializer):
    file = ChecksumFileSerializer()
    vtp_data = ChecksumFileSerializer(required=False)

    def to_representation(self, value):
        ret = super().to_representation(value)
        return ret

    class Meta:
        model = models.Mesh3D
        fields = '__all__'


class Mesh3DDataSerializer(Mesh3DSerializer):
    def to_representation(self, value):
        ret = super().to_representation(value)
        # Extract data as base64
        with value.vtp_data.yield_local_path() as path:
            with open(path, 'rb') as data:
                data_content = data.read()
                base64_content = base64.b64encode(data_content)
                base64_content = base64_content.decode().replace('\n', '')
        ret['vtp_data'] = base64_content
        return ret


class Tiles3DSerializer(serializers.ModelSerializer):
    json_file = ChecksumFileSerializer()

    class Meta:
        model = models.Tiles3D
        fields = '__all__'


class Tiles3DMetaSerializer(SpatialEntryFootprintSerializer):
    source = Tiles3DSerializer(read_only=True)

    class Meta:
        model = models.Tiles3DMeta
        fields = '__all__'



import logging
import os
from pathlib import Path

from django.contrib.gis.db import models
from django.core.exceptions import ValidationError
from django_extensions.db.models import TimeStampedModel
from rgd.models import ChecksumFile, FileSet, SpatialEntry
from rgd.models.mixins import DetailViewMixin, TaskEventMixin
from rgd.models.utils import get_or_create_checksumfile_file_field
from rgd_3d.tasks import jobs

logger = logging.getLogger(__name__)


class Tiles3D(TimeStampedModel, TaskEventMixin, DetailViewMixin):
    class Meta:
        verbose_name = '3D tiles'
        verbose_name_plural = '3D tiles'

    def clean(self):
        if self.json_file.file_set is None:
            raise ValidationError('"json_file" must be part of a FileSet.')
        return super().clean()

    name = models.CharField(max_length=1000, blank=True)
    description = models.TextField(null=True, blank=True)

    json_file = models.ForeignKey(
        ChecksumFile,
        on_delete=models.CASCADE,
        related_name='+',
        help_text='The `tileset.json` file that contains metadata about this set of 3D tiles.',
    )

    task_funcs = (jobs.task_read_3d_tiles_file,)
    detail_view_name = 'detail-tiles-3d'


class Tiles3DMeta(TimeStampedModel, SpatialEntry):
    source = models.OneToOneField(Tiles3D, on_delete=models.CASCADE)

    @property
    def name(self):
        return self.source.json_file.name

    detail_view_name = Tiles3D.detail_view_name
    detail_view_pk = 'source__pk'


def create_tiles3d_from_paths(paths, collection=None):
    # Start by ingesting tileset.json
    # tileset_json_path.resolve().parent
    tileset_json_path = Path([path for path in paths if path.split('/')[-1] == 'tileset.json'][0])
    tileset_json_name = '/'.join(str(tileset_json_path).split('/')[-2:])

    logger.error(tileset_json_path)

    dataset_name = os.path.basename(tileset_json_path.parent)

    # Create a FileSet for the 3D tiles dataset
    fileset, created = FileSet.objects.get_or_create(name=dataset_name)

    with open(tileset_json_path, 'rb') as f:
        tileset_json_checksum_file, _ = get_or_create_checksumfile_file_field(
            f, collection=collection
        )
    tileset_json_checksum_file.file_set = fileset
    tileset_json_checksum_file.name = tileset_json_name
    tileset_json_checksum_file.save(update_fields=['file_set', 'name'])

    # Filter out tileset.json as we have already ingested that
    paths = [path for path in paths if path.split('/')[-1] != 'tileset.json']

    for path in paths:
        try:
            Path(path).relative_to(tileset_json_path.parent)
        except ValueError:
            continue
        # Strip off irrelevant parts of file path
        logger.error(path)
        file_name = Path(path).relative_to(tileset_json_path.parent.parent)
        logger.error(file_name)
        with open(path, 'rb') as f:
            checksum_file, _ = get_or_create_checksumfile_file_field(f, collection=collection)
        checksum_file.file_set = fileset
        checksum_file.name = file_name
        checksum_file.save(update_fields=['file_set', 'name'])

    entry, created = Tiles3D.objects.get_or_create(
        name=dataset_name, json_file=tileset_json_checksum_file
    )
    return entry



from django.db import transaction
from django.db.models.base import ModelBase
from django.db.models.signals import post_save
from django.dispatch import receiver
from rgd.utility import skip_signal
from rgd_3d import models


@receiver(post_save, sender=models.Mesh3D)
@skip_signal()
def _post_save_mesh_3d_file(sender: ModelBase, instance: models.Mesh3D, *args, **kwargs):
    transaction.on_commit(lambda: instance._post_save_event_task(*args, **kwargs))


@receiver(post_save, sender=models.Tiles3D)
@skip_signal()
def _post_save_tiles_3d_file(sender: ModelBase, instance: models.Tiles3D, *args, **kwargs):
    transaction.on_commit(lambda: instance._post_save_event_task(*args, **kwargs))
